/*******************************************************************************
* Copyright (c) 1993, 1995 by Charles Kooperberg.                              *
* All rights reserved. This function is part of the Logspline Density          *
* Estimation Program. It was written by Charles Kooperberg at the University   *
* of Washington and the University of California at Berkeley between 1988 and  *
* 1991. It is described in the paper: Kooperberg, Charles and Stone, Charles J.*
* `Logspline density estimation for censored data', Journal of Computational   *
* and Graphical Statistics, 1 (1992) 301-328.                                  *
* You are free to use this program, for non-commercial purposes only,          *
* under two conditions:                                                        *
* (1) This note is not to be removed                                           *
* (2) Publications using logspline computations should reference the           *
*  publication  mentioned above.                                               *
* For questions, please email clk@stat.washington.edu                          *
*                       Charles Kooperberg, November 30, 1995                  *
*******************************************************************************/
/* this file contains functions that are executed - before the iterations start-
   by the main program.

   coeff  - computes the coefficient matrices for the basis functions          
   start1 - gathers statistics used to compute the starting values
   start2 - computes the starting values
   sufficient1 - gathers statistics used to compute the sufficientt statistics
   sufficient2 - computes the sufficientt statistics
   knotplace   - places the knots
   knotnumber  - determines the number of starting knots
   piecedens   - determines a piecewise constant density for initial knot placem
*******************************************************************************/
#include "mhead.h"

void coeff(coef2)

double coef2[][NC];
/* This function computes the coefficients of the basis functions from the knots
   coef2:
   first index: basis function number-1,
   second index: 0:1, 1:x, 2:(x-t1)+^3,  3:(x-t2)+^3, 4:(x-t3)+^3,........
   coef: 
   between knot(i) and knot(i+1) the coefficient of x^power of basisfunction(j)
   first index: basis function number-1 (j-1)
   second index: power of x
   third index: interval (i)                                                  

   Basis functions B(2)-B(nknots-3) are multiples of B-splines.
   Further the coefficients are choosen such that the quadratic and  
   cubic terms in both tails are 0;  this leads to differnt basis
   functions for B(1), B(nknots-2) and B(nknots-1). B(1) is linear left
   of the first knot. B(nknots-2) is constant to the right of the last
   knot and B(nknots-1) is linear to the right of the last knot.              */


{
   int i,j,k;
   double z0,z1;
/* local
   i j k       - counter
   z0,z1       - value of constants of two succesive basisfunctions           */

/* Initializations                                                            */
   for(i=0; i<nknots; i++){
      for(j=0; j<nknots+3; j++){
         coef2[i][j]=0.;
         if(j<nknots+2)
            for(k=0; k<4; k++)
               coef[i][k][j]=0.;
      }
   }

/* The coefficients for the two tail basis functions are easy to compute      */
   coef2[0][2] = 1.;
   coef2[0][3] = (knots[0] - knots[2]) / (knots[2] - knots[1]);
   coef2[0][4] = (knots[1] - knots[0]) / (knots[2] - knots[1]);
   coef2[0][1] = -3. * (pow(knots[0], 2.) 
                     + coef2[0][3] * pow(knots[1], 2.)
                     + coef2[0][4] * pow(knots[2], 2.));
   coef2[0][0] = - knots[nknots-1] * coef2[0][1]
                 - coef2[0][2] * pow((knots[nknots-1] - knots[0]), 3.)
                 - coef2[0][3] * pow((knots[nknots-1] - knots[1]), 3.)
                 - coef2[0][4] * pow((knots[nknots-1] - knots[2]), 3.);
   coef2[0][5] = 0.;
   coef2[nknots-2][nknots-1] = 1.;
   coef2[nknots-2][nknots]   = (knots[nknots-3] - knots[nknots-1]) / 
                               (knots[nknots-1] - knots[nknots-2]);
   coef2[nknots-2][nknots+1] = (knots[nknots-2] - knots[nknots-3]) /
                               (knots[nknots-1] - knots[nknots-2]);

/* we first create basis functions that are 0 before knot[i] and constant
   after knot [i+3]                                                           */

   if(nknots >= 4){
      for(i=1; i<nknots-2; i++){
         coef2[i][i+1] = 1.;
         coef2[i][i+4] = (knots[i+1]-knots[i-1]) * (knots[i-1]-knots[i])
                     / ((knots[i+1]-knots[i+2]) * (knots[i]-knots[i+2]));
         coef2[i][i+3] = (coef2[i][i+4] * (knots[i]-knots[i+2])
                     + knots[i] - knots[i-1]) / (knots[i+1]-knots[i]);
         coef2[i][i+2] = -1. - coef2[i][i+3] - coef2[i][i+4];
      }

/* In the following part we subtract a number of times one basis
   function from another - so that basis function i becomes 0 after knot[i+4] */
      if(nknots >= 5){
         for(i=1; i<nknots-3; i++){
            z0 = 0.;
            z1 = 0.;
            for(j=2; j<nknots+1; j++){
               z0 = z0 + coef2[i][j]   * pow((knots[nknots-1]- knots[j-2]), 3.);
               z1 = z1 + coef2[i+1][j] * pow((knots[nknots-1]- knots[j-2]), 3.);
            }
            for(j=2; j<nknots+2; j++)
                  coef2[i][j] = coef2[i][j] - (z0 / z1) * coef2[i+1][j];
         }
      }
   }

/* Now the coef matrix. First basis function 1.                               */

   for(k=0; k<3; k++){
      coef[0][1][k] = coef2[0][1];
      coef[0][0][k] = coef2[0][0];
   }

/* The rest is a bit tricking with the correct indices                        */
   for(i=0; i<nknots-1; i++){
      for(j=i; j<i+4; j++){
         if(j > 0 && j < nknots+1 && (i != 0 || j != 3)){
            for(k=i+1; k<j+2; k++){
               if(k != 1){
                  coef[i][0][j] = coef[i][0][j] - 
                                  coef2[i][k] * pow(knots[k-2], 3.);
                  coef[i][1][j] = coef[i][1][j] + 
                                  3. * coef2[i][k] * pow(knots[k-2], 2.);
                  coef[i][2][j] = coef[i][2][j] - 3. * coef2[i][k] * knots[k-2];
                  coef[i][3][j] = coef[i][3][j] + coef2[i][k];
               }
            }
         }
      }
   }
}
/******************************************************************************/

/* These are two functions to compute the starting values.
   The first one gathers statistics from all the datapoints and knots.
   The second one computes the starting values from these statistics          */

void start1(crossprods,derivatives,sample,nsample)

/* the objective of this routine is the computation of derivatives and
   crossproducts                                                              */

double crossprods[][NC],derivatives[],sample[];
int nsample[];

/* sample,nsample,nknots, accuracy and knots see lhead.h and originating file
   derivatives[0]   = sum((sample))'' which is 0....
   derivatives[i]   = sum(([(sample-knot[i-1])+]^3)'')
   crossprods[0][0] = sum((sample)'*(sample)')=nsample[0]
   crossprods[i][0] = sum(([(sample-knot[i-1])+]^3)'*(sample)')
   crossprods[0][i] = sum(([(sample-knot[i-1])+]^3)'*(sample)')
   crossprods[i][j] = sum(([(sample-knot[i-1])+]^3)'*([(sample-knot[j-1])+]^3)')
   they are used in start2 to compute starting values                         */

{
   int i,j,k,i2,i3;
   double xs,uuu,rr,rs;
/* counters, utility                                                          */

/* Initializations                                                            */
   for(i=0; i<nknots+1; i++){
      derivatives[i] = 0.;
      for(j=0; j<nknots+1; j++)
         crossprods[i][j] = 0.;
   }
   crossprods[0][0] = nsample[0];

   rr=sample[0];
   rs=sample[0];
/* First the uncensored part                                                  */
   for(i=0;i<nsample[0]+nsample[2];i++) if(sample[i]>rr) rr=sample[i];
   for(i=0;i<nsample[0]+nsample[2];i++) if(sample[i]<rs) rs=sample[i];
   if(nsample[1]>0)for(j=0; j<nknots; j++){
      for(i=0;i<nsample[1];i++){
         if(sample[i]>knots[j]){
            derivatives[j+1] = derivatives[j+1] + 6 * (sample[i] - knots[j]);
            crossprods[j+1][0]=crossprods[j+1][0]+3*pow(sample[i]-knots[j],2.);
            for(k=0; k<=j; k++)
               crossprods[j+1][k+1] = crossprods[j+1][k+1] + 
                  9 * pow(((sample[i] - knots[j]) * (sample[i] - knots[k])),2.);
         }
      }
   }
/* The interval censored part - we take the midpoints of the intervals        */
   if(nsample[2]>0)for(j=0; j<nknots; j++){
      for(i=0;i<nsample[2];i++){
         xs=(sample[nsample[1]+2*i]+sample[nsample[1]+2*i+1])/2.;
         if(xs>knots[j]){
            derivatives[j+1] = derivatives[j+1] + 6 * (xs - knots[j]);
            crossprods[j+1][0]=crossprods[j+1][0]+3*pow(xs-knots[j],2.);
            for(k=0; k<=j; k++)
               crossprods[j+1][k+1] = crossprods[j+1][k+1] + 
                  9 * pow(((xs - knots[j]) * (xs - knots[k])),2.);
         }
      }
   }
/* The right censored part                                                    */
   if(nsample[3]>0)for(j=0; j<nknots; j++){
      for(i2=0;i2<nsample[3];i2++){
         i=nsample[1]+2*nsample[2]+i2;
         for(i3=0;i3<4;i3++)uuu=(double)(i3)*rr/4.+(4-i3)*sample[i]/4.;
         if(uuu>knots[j]){
            derivatives[j+1] = derivatives[j+1] + 1.5 * (uuu - knots[j]);
            crossprods[j+1][0]=crossprods[j+1][0]+0.75*pow(uuu-knots[j],2.);
            for(k=0; k<=j; k++)
               crossprods[j+1][k+1] = crossprods[j+1][k+1] + 
                  2.25 * pow(((uuu - knots[j]) * (uuu - knots[k])),2.);
         }
      }
   }

/* The left  censored part                                                    */
   if(nsample[4]>0)for(j=0; j<nknots; j++){
      for(i2=0;i2<nsample[4];i2++){
         i=nsample[1]+2*nsample[2]+nsample[3]+i2;
         for(i3=0;i3<4;i3++)uuu=(double)(i3)*rs/4.+(4-i3)*sample[i]/4.;
         if(uuu>knots[j]){
            derivatives[j+1] = derivatives[j+1] + 1.5 * (uuu - knots[j]);
            crossprods[j+1][0]=crossprods[j+1][0]+0.75*pow(uuu-knots[j],2.);
            for(k=0; k<=j; k++)
               crossprods[j+1][k+1] = crossprods[j+1][k+1] + 
                  2.25 * pow(((uuu - knots[j]) * (uuu - knots[k])),2.);
         }
      }
   }

/* symmetarize crossprods                                                     */
   for(j=0; j<nknots; j++)
      for(k=j+1; k<nknots+1; k++)
         crossprods[j][k] = crossprods[k][j];
   return;
}

/******************************************************************************/
void start2(crossprods,derivatives,coef2,nkstart,iremove)

/* this function combines derivatives and crossprods, and then inverts a
   system to compute starting values                                          */

double crossprods[][NC],derivatives[],coef2[][NC];
int nkstart,iremove;

{
   int i,j,k,l,j2,l2,infox,kpvt[200];
   double work[NC][NC],r1,r2,r3,r4;

/* local:
   i j j2 k l l2  - counters
   work           - workmatrix, 
   infox kpvt     - reguired by inversion routine                             */

/* Initializations */
   r1=theta[0];
   r2=theta[1];
   r3=theta[nknots-2];
   r4=theta[nknots-1];
   if(nknots>4){
      for(i=0; i<nknots; i++){
         theta[i] = 0.;
         for(j=0; j<nknots; j++)
            work[i][j] = 0.;
      }

/* The left hand side is a matrix with sums of crossproducts of derivatives of
   basis functions.  We make them by combining coefficients (coef2) and
   crossproducts (crossprods) of the powerbasis.  The right hand side is a
   vector with second derivatives of the basis functions.  Again combining
   coefficients (coef2) now with derivatives.                                 */

      for(i=0; i<nknots-1; i++){
/* these are the only combinations for which coef2 is not equal to 0          */
         for(j2=i+1; (j2<i+6 && j2<nknots+2); j2++){
            j = j2;
            if(i == 0 && j == 5)j = 1;
            theta[i] = theta[i] - coef2[i][j] * derivatives[j-1];
            for(k=i; k<nknots-1; k++){

/* these are the only combinations for which coef2 is not equal to 0          */
               for(l2=k+1; (l2<k+6 && l2<nknots+2); l2++){
                  l = l2;
                  if(k == 0 && l == 5)l = 1;
                  work[i][k] = work[i][k] +
                     coef2[i][j] * coef2[k][l] * crossprods[j-1][l-1];
               }
            }
         }
      }
      
/* Make symmetric                                                             */
      for(i=1; i<nknots-1; i++)
         for(j=0; j<i; j++)
            work[i][j]=work[j][i];

/* Solving the system finishes the job.                                       */
      i=nknots-1;
      j=NC;
      xdsifa_(work,&j,&i,kpvt,&infox);
      xdsisl_(work,&j,&i,kpvt,theta);
/* If we had previous iterations - we're sometimes better of using that       */
      if(nkstart!=nknots){
         if(iremove>2){
            theta[0]=r1;
            theta[1]=r2;
         }
         if(iremove<3){
            theta[0]=r1/25.;
            theta[1]=r2/25.;
         }
         if(iremove<nknots-2){
            theta[nknots-3]=r3;
            theta[nknots-2]=r4;
         }
         if(iremove>nknots-3){
            theta[nknots-3]=r3/25.;
            if(nknots==4&&iremove==3)theta[1]=(r2+r3)/50.;
            theta[nknots-2]=r4/25.;

         }
      }
/* If the thetas are too large, we are better of if we make them smaller      */
      else{
         r1=0.;
         for(i=0;i<nknots-2;i++)r1=r1+fabs(theta[i]);
         if(r1>10000)
            for(i=0;i<nknots-2;i++)theta[i]=theta[i]*10000/r1;
      }
   }
   else{
      for(i=0;i<3;i++) theta[i]=0.;
   }
}
/******************************************************************************/

/* The first routine collects statistics about the data. The second one
   uses these to compute the sufficiemt statistics                            */

void suffstat1(suffcombine,sample,nsample)

double suffcombine[][2],sample[];
int nsample[];
/* these quantities are defined in lhead.h and the file where they originate  
   suffcombine[0][a] = 1
   suffcombine[1][a] = sum(sample)/nsample[0]
   suffcombine[i][a] = sum(((sample-knot[i-1])+)^3)/nsample[0]                    
   where a is 0 for the exact sample and 1 for the interval censored stuff.
   they are later (in suffstat2) combined to compute the sufficient statistics*/

{
   int i,j;
   double rr;
/* counters, utility                                                          */

/* initialization */
   for(i=1;i<nknots+2;i++){
      suffcombine[i][0] = 0.;
      suffcombine[i][1] = 0.;
   }

   suffcombine[0][0] = (double)nsample[1]/(double)nsample[0];
   suffcombine[0][1] = (double)nsample[2]/(double)nsample[0];

/* for the complete observations                                              */
   for(j=0; j<nsample[1]; j++){
      suffcombine[1][0] = suffcombine[1][0] + sample[j]/(double)nsample[0];
      for(i=2; i<nknots+2; i++)
         if(sample[j] > knots[i-2])
           suffcombine[i][0] = suffcombine[i][0] + 
                           pow(sample[j]  - knots[i-2], 3.)/(double)nsample[0] ;
   }
/* for the interval censored observations                                     */
   for(j=0; j<nsample[2]; j++){
      rr=(sample[nsample[1]+2*j]+sample[nsample[1]+2*j+1])/2.;
      suffcombine[1][1] = suffcombine[1][1] + rr/(double)nsample[0];
      for(i=2; i<nknots+2; i++)
         if(rr > knots[i-2])
           suffcombine[i][1] = suffcombine[i][1] + 
                          pow(rr  - knots[i-2], 3.)/(double)nsample[0] ;
   }
}

/******************************************************************************/

void suffstat2(suffcombine,coef2,sufficient)

double suffcombine[][2],coef2[][NC],sufficient[][2];
/* all defined in lhead.h and the file where they originate.
   suffcombine defined in suffstat1                                           */

{
   int i,j;
/* counters                                                                   */

/* over the basisfunctions                                                    */
   for(i=0;i<nknots;i++){
      sufficient[i][0] = 0.;
      sufficient[i][1] = 0.;
         
/* over the intervals                                                         */
      for(j=0;j<nknots+2;j++){
         sufficient[i][0] = sufficient[i][0] + suffcombine[j][0] * coef2[i][j];
         sufficient[i][1] = sufficient[i][1] + suffcombine[j][1] * coef2[i][j];
      }
   }
}
/******************************************************************************/
/* These routines determine the position of the knots and the number of knots.*/

void knotplace(iknots,rknots,iknotauto,bound,sample,nsample,SorC,smp2,smp3,qt)
                                    

int iknots[],iknotauto,SorC[],nsample[];
double rknots[],sample[],bound[],smp2[],smp3[],qt[];
/* these quantities are defined in lhead.h and the files where they originate */

{
   int i,j,j2,k,kk,ll,ia,il,piecedens();
/* local integers
   i k     - counters
   j j2    - is there an odd or an even number of knots?                      */

   double s,eps,eps1,eps2,v,w,five=4,u1,u2,u3,u4;
/* local   - double precision
   five    - parameter for knot rule
   s       - this parameter has to become (nsample[0]+1)/2
   eps     - current estimate for five rule
   eps1    - lower bound for eps
   eps2    - upper bound for eps
   v w     - combinations of counters                                         */

/* the major part of this function is computing the knots for the
   five rule. The essence is to find eps such that s=(nsample[0]-1)/2
   The usual make it a bit larger, smaller, till we have one just
   too large and one just too small.                                          */

   qt[0]=-1.;
   qt[1]=1.;
   if(iknotauto == 0&&nknots>12){
      qt[0]=knots[4];
      qt[1]=knots[nknots-5];
   }
   if(iknotauto == 0&&nknots>9 && nknots <13){
      qt[0]=knots[3];
      qt[1]=knots[nknots-4];
   }
   if(iknotauto == 0&&nknots>6 && nknots <10){
      qt[0]=knots[2];
      qt[1]=knots[nknots-3];
   }
   if(iknotauto == 0&&nknots==3){
      qt[0]=(knots[0]+knots[1])/2.;
      qt[1]=(knots[2]+knots[1])/2.;
   }
   if(iknotauto == 0&&nknots>3 &&nknots<7){
      qt[0]=knots[1];
      qt[1]=knots[nknots-2];
   }
   if(iknotauto == 1){

/* compute the piecewise density                                              */
      il = piecedens(sample,smp2,smp3,nsample);
      kk = floor((double)il/2+2.1);
      if(nknots>kk && nsample[5]==0){
          nknots=kk;
          if(SorC[0] == 0)
               printf("running with maximum number of degrees of freedom");
          else
             SorC[20] = 1;
      }
      if(nknots==nsample[5]+1)SorC[20]=1;

/* Check whether there are not too many knots                                 */
      kk=0;
      ll=0;
      do{
         ll=kk+ll;
         kk=0;
         if(bound[1] < 0.5 && bound[3] < 0.5){

/* all knots are a minimum of "five" apart                                    */

            if((nknots-1) * five >= nsample[0] - 1){
               i = floor(1. + (nsample[0] - 1.) / five);
               if(SorC[0] == 0){
                  printf("too many knots, at most %d knots possible\n",i);
                  SorC[0] = -647;
               }
               else {
                  SorC[0] = -2;
                  SorC[1] = i;
               }
               return;
            }

/*  place the 2 extreme knots                                                 */

            rknots[0] = 1.;
            rknots[nknots-1] = nsample[0];

/*  j and j2 are this way to deal both with odd and even situations           */

            j = ceil((nknots-1)/2.);
            j2 = floor((nknots-1)/2.);

/*  eps1 and eps2 are lower and upper bound on eps, eps is our first guess.   */

            eps1 = five - pow(((nsample[0] - 1) / five),(1. / (j - 1)));
            if(eps1 > 0.)
               eps1 = 0.;
            eps2 = five - 1.;

/* s should become exactly (nsample[0]-1)/2: in that case the knots
   are symmetric and cover exactly the whole range.
   We here compute what s would be for the present value of eps 
   s is the span of all the knots: i.e. the location of the middle knot
   if the first knot is at 1, nknots as specified and eps as guessed.         */

            do{
               eps = (eps1 + eps2) / 2.;
               s = 1.;
               w = five;
               for(i=1; i<=j2; i++){
                  v = i;
/* s is the location after adding another knot                                */
                  s = s + w;   
/* we store the rknots - in case they are good                                */
                  rknots[i] = s;
                  rknots[nknots-i-1] = nsample[0] + 1 - s;
                  v = five - v * eps;
/* w is what is going to be the next gap                                      */
                  if(v < 1.) v = 1.;
                  w = w * v;
               }
/* Are there an odd or even number of gaps?                                   */
               if(j * 2 == nknots) 
/* even - no more knots to place.                                             */
                  s = s + w/2.;     
               else
/* odd - the last knot                                                        */
                  rknots[j] = (nsample[0] + 1.) / 2.;
   
/* Is eps too large or too small?                                             */
               if(2. * s >= nsample[0]+1) 
                  eps1 = eps;
               else
                  eps2 = eps;
   
/* Are eps1 and eps2 close together                                           */
            } while(eps2 - eps1 > .001);   
         }
         else{
/* if both sides there are finite limits we put them equidistant              */
            if(bound[1] > 0.5 && bound[3] > 0.5){
               for(i=0;i<nknots;i++)rknots[i]=
                   1.+(double)(nsample[0]-1)*((double)i)/(double)(nknots-1);
            }
            else{
/* one side bounded, one side unbounded half of the five-rule              
   which side is bounded?                                                     */
               rknots[0]=1.;
               rknots[nknots-1]=nsample[0];
               eps1 = 0.;
               eps2 = five;
               do {
                  eps = (eps1 + eps2)/2.;
                  s = 1.;
                  w = five;
                  for(i=1;i<nknots-1;i++){
                     v = i;
/* s is the location after adding another knot                                */
                     s = s+w;
/* we store the rknots - in case they are good                                */
                     if(bound[3] > 0.5)
                        rknots[i] = s;
                     else
                        rknots[nknots-i-1]=nsample[0]+1.-s;
                     v = five - v *eps;
/* w is what is going to be the next gap                                      */
                     if(v < 1.) v = 1.;
                     w = w * v;
                 }
/* Is eps too large or too small?                                             */
                 if(s + w >(double)nsample[0])
                    eps1 = eps;
                 else
                    eps2 = eps;
/* Are eps1 and eps2 close together                                           */
              } while(eps2 - eps1 > .001);
            }
         }
         if(nknots==3)rknots[1]=nsample[0]/2.;

/* Translate rknots in knots                                                  */
         if(ll==0){
/* the first two knots are easy                                               */
            knots[0] = smp2[0];     
            knots[nknots-1] = smp2[il];

 /* cycle through the endpoints, average                                      */
            k = 0;
            for(i=1;i<nknots-1;i++){
               rknots[i]=(rknots[i]-0.5);
               for(k=k;k<=il;k++){
                  if(smp3[k]>=rknots[i]){
                     knots[i]=((rknots[i]-smp3[k-1])*smp2[k]+
                             (smp3[k]-rknots[i])*smp2[k-1])/(smp3[k]-smp3[k-1]);
                     ia=k;
                     k=k+4*il;
                  }
               }
               k=ia;
            }
         }

/* find the first and the third quartile                                      */
         for(i=1;i<il;i++){
            if(smp3[i]>0.25*nsample[0]){
               u3=smp2[i];  
               i=i+2*il;
               qt[0]=u3;
            }
         }
         for(i=1;i<il+1;i++){
            if(smp3[i]>0.75*nsample[0]){
               u4=smp2[i];  
               i=i+2*il;
               qt[1]=u4;
            }
         }

/* knots that are close together at the end, are an indicator of a discontinuity
   we replace the knots if this happens. kk is an indicator that somthing like
   this is the case - we fix this by, temporarily, cheating and saying that
   there is a discontinuity (changing bound), ll reminds us to change back, by
   the way, u2 is the IQR of the data                                         */

         if(bound[1]<0.5){
            u1=knots[2]-knots[0];
            u2=u4-u3;
            if(u1<u2/30.){
               if(SorC[0]==0){
                  printf("possible discontinuity of the density at left end\n");
                  printf("considder rerunning it with lbound=min(sample)\n");
               }
               else SorC[21]=1;
               kk=kk+1;
               bound[1]=1;
               bound[2]=knots[0];
            }
         }
         if(bound[3]<0.5){
            u1=knots[nknots-1]-knots[nknots-3];
            u2=u4-u3;
            if(u1<u2/30.){
               if(SorC[0]==0){
                 printf("possible discontinuity of the density at right end\n");
                  printf("considder rerunning it with ubound=max(sample)\n");
               }
               else SorC[22]=1;
               kk=kk+2;
               bound[3]=1;
               bound[4]=knots[nknots-1];
            }
         }
      }while(kk>0);
      if(ll==1||ll==3)bound[1]=0.;
      if(ll==2||ll==3)bound[3]=0.;
   
/* if after previous changes these knots are still to close together, the
   density might be real high at the end - we toss out knots in this case     */
   u1=knots[nknots-1]-knots[nknots-3];
   u2=u4-u3;
   if(u1<u2/30){
      if(SorC[0]==0){
      printf("potential very high value of the density at right end\n");
      printf("running program with fewer knots than usual\n");
      }
      else SorC[22]=3;
      knots[nknots-3]=(knots[nknots-3]+knots[nknots-2])/2.;
      knots[nknots-2]=knots[nknots-1];
      nknots=nknots-1;
   }
   do{
      u1=knots[nknots-1]-knots[nknots-3];
      u2=u4-u3;
      if(u1<u2/30){
         knots[nknots-3]=(knots[nknots-3]+knots[nknots-2])/2.;
         nknots=nknots-1;
      }
   }while(u1<u2/30);
   u1=knots[2]-knots[0];
   u2=u4-u3;
   if(u1<u2/30){
      if(SorC[0]==0){
         printf("potential very high value of the density at left end\n");
         printf("running program with fewer knots than usual\n");
      }
      else SorC[21]=3;
      knots[1]=(knots[1]+knots[2])/2.;
      for(i=2;i<nknots-1;i++)knots[i]=knots[i+1];
      nknots=nknots-1;
   }
   do{
      u1=knots[2]-knots[0];
      u2=u4-u3;
      if(u1<u2/30){
         knots[1]=(knots[1]+knots[2])/2.;
         for(i=1;i<nknots-1;i++)knots[i]=knots[i+1];
         nknots=nknots-1;
      }
   }while(u1<u2/30);
   }

/* Remove double knots,either store error messages, or print them.            */

   j = 0;                          
   k = 0;
   for(i=1; i<nknots; i++){
      if(knots[i] > knots[j]){
         j++;
         knots[j]=knots[i];
      }
      else {
         k++;
         if(SorC[0] == 8){
            printf("===> warning: knot %d removed - double knot\n",i+1);
            if(k == 2){
               printf("* several double knots suggests that your data is *\n");
               printf("* strongly rounded: attention might be required.  *\n");
            }
         }
         else {
            SorC[0] = 2 + k;
            if(SorC[0] == 23)return;
            SorC[k] = i+1;
         }
      }
   }
   nknots = j+1;
   u3=2./(qt[1]-qt[0]);
   u4=1.-2.*qt[1]/(qt[1]-qt[0]);
   qt[0]=u4;
   qt[1]=u3;
   for(i=0;i<nknots;i++){
      rknots[i] = u4 + u3 * knots[i];
      knots[i] = u4 + u3 * knots[i];
      iknots[i] = i;
   }           
   for(i=0;i<nsample[0]+nsample[2];i++)sample[i]=u4+u3*sample[i];
}

/******************************************************************************/
/* determines the number of (starting) knots                                  */

int knotnumber(idelete,nsample,nknots,SorC)

int idelete,nsample[],nknots,SorC[];
/* all defined in lhead.c and originating file                                */
{

   int nn;
   double r;

   nn=nsample[1]+nsample[2]+(nsample[3]+nsample[4])/2;
   if(nknots>=NC || (nknots > nsample[5]+1 && nsample[5]>1)){ 
      if(SorC[0]==0)
         printf("can not run with that many knots\n");
      else
         SorC[20]=1;
      nknots=NC-1;
      if(nknots > nsample[5]+1)nknots=nsample[5]+1;
   }
   if(nknots > 0)
    return nknots;
   r = 2.5*pow((double)nn,0.2);
   if(idelete > 0)r=1.6*r;
   if(idelete > 0 && nn < 51)r=(double)nn/6.;
   if(r>NC-1.1)r=NC-1.1;
   if(r<2.5)r=2.5;
   if(nsample[5]!=0 && r>nsample[5]+1)r=nsample[5]+0.1;
   return ceil(r);
}
/******************************************************************************/
int piecedens(sample,smp2,smp3,nsample)
double smp2[],smp3[],sample[];
int nsample [];
{
   int i,j,il,k,m,n;
   if(nsample[1]>0){
      i=1;
      j=nsample[1];
      xssort_(sample,smp3,&j,&i);
   }
   if(nsample[2]>0){
      for(i=0;i<nsample[2];i++){
         smp2[i]=sample[nsample[1]+2*i];
         smp3[i]=sample[nsample[1]+2*i+1];
      }
      i=2;
      j=nsample[2];
      xssort_(smp2,smp3,&j,&i);
      for(i=0;i<nsample[2];i++){
         sample[nsample[1]+2*i]=smp2[i];
         sample[nsample[1]+2*i+1]=smp3[i];
      }
   }
   if(nsample[3]>0){
      for(i=0;i<nsample[3];i++) smp2[i]=sample[i+nsample[1]+2*nsample[2]];
      i=1;
      j=nsample[3];
      xssort_(smp2,smp3,&j,&i);
      for(i=0;i<nsample[3];i++) sample[i+nsample[1]+2*nsample[2]]=smp2[i];
   }
   if(nsample[4]>0){
      for(i=0;i<nsample[4];i++)
         smp2[i]=sample[i+nsample[1]+2*nsample[2]+nsample[3]];
      i=1;
      j=nsample[4];
      xssort_(smp2,smp3,&j,&i);
      for(i=0;i<nsample[4];i++)
         sample[i+nsample[1]+2*nsample[2]+nsample[3]]=smp2[i];
   } 
   for(i=0;i<nsample[1];i++){
      smp2[i]=sample[i];
      smp3[i]=1.;
   }
   for(i=0;i<nsample[3]+nsample[4];i++){
      j=nsample[1]+2*nsample[2]+i;
      smp2[j]=sample[j];
      smp3[j]=1.;
   }
   if(nsample[2]>0){
      for(i=0;i<nsample[2];i++){
         j=nsample[1]+i*2;
         smp2[j]=sample[j];
         smp3[j]=0.;
         smp2[j+1]=sample[j+1];
         smp3[j+1]=0.;
      }
   }
   i=2;
   j=nsample[0]+nsample[2];
   xssort_(smp2,smp3,&j,&i);
   i=0;
   for(j=1;j<nsample[0]+nsample[2];j++){
      if(smp2[j]==smp2[i])
         smp3[i]=smp3[i]+smp3[j];
      else{
         i++;
         smp2[i]=smp2[j];
         smp3[i]=smp3[j];
      }
   }
   il=i;
      
   for(i=il-1;i>=0;i=i-1){
      if(smp3[i]>0){
         smp3[i+1]=smp3[i+1]+0.5*smp3[i];
         smp3[i]=0.5*smp3[i];
      }
   }
   if(nsample[2]>0){
      k=0;
      for(i=0;i<nsample[2];i++){
         j=nsample[1]+i*2;
         for(m=k;m<=il;m++){
            if(sample[j]<smp2[m]){
               k=m;
               m=2*il+1;
               for(n=k;n<=il;n++){
                  if(sample[j+1]<smp2[n])n=2*il+1;
                  else
                  smp3[n]=smp3[n]+(smp2[n]-smp2[n-1])/(sample[j+1]-sample[j]);
               }
            }
         }
      }
   }
   for(i=1;i<=il;i++) smp3[i]=smp3[i]+smp3[i-1];
   return il;
}
