/*******************************************************************************
* Copyright (c) 1993, 1995 by Charles Kooperberg.                              *
* All rights reserved. This function is part of the Logspline Density          *
* Estimation Program. It was written by Charles Kooperberg at the University   *
* of Washington and the University of California at Berkeley between 1988 and  *
* 1991. It is described in the paper: Kooperberg, Charles and Stone, Charles J.*
* `Logspline density estimation for censored data', Journal of Computational   *
* and Graphical Statistics, 1 (1992) 301-328.                                  *
* You are free to use this program, for non-commercial purposes only,          *
* under two conditions:                                                        *
* (1) This note is not to be removed                                           *
* (2) Publications using logspline computations should reference the           *
*  publication  mentioned above.                                               *
* For questions, please email clk@stat.washington.edu                          *
*                       Charles Kooperberg, November 30, 1995                  *
*******************************************************************************/
/* this file contains the following functions:
expin  - computes an exponential integral
expin2 - computes another exponential integral
dens3  - computes a logspline-density in one point
dens33 - computes the log of a logspline-density in one point
numint - computes a numerical integral
numints- computes a vector of numerical integrals                             */

/******************************************************************************/
/*   this function computes (analytical) the integral:   
                                                          
      t1 t1 inf t2                                         
      /   /  /  /     2.          a4*x+a5                       
      |   |  |  |(a1*x +a2*x+a3)*e       dx                        
      /   /  /  /                                            
    -inf t2 t1 t1                                              
      1   2.  3  4     <==== version                              */
                                                                 
#include "mhead.h"

double expin(version,t1,t2,a)

/* input: a,b,c,d,e,t1,t2,version: see figure above
   local: a1,b1,c1: as a,b and c, but for the primitive.                      
          f1,f2 half-products                                    */
int version;
double t1,t2,a[];
/* this version does not contain much information - the best way to figure out
   what is happening is to compute the integrals above, and then check below  */

{
   double a1,b1,c1,f1,f2;
   int i1,i2;
/* if d=0 and version is 1 or 3 well...........                               */
   if(a[4]!=0 || version == 1 || version ==3){
      a1 = a[1]/a[4];
      b1 = (a[2]-2*a1)/a[4];
      c1 = (a[3]-b1)/a[4];
      f1 = a1*t1*t1+b1*t1+c1; 
      i1 = 1;
      if(f1<0) i1 = -1;
      f1 = log(fabs(f1)) + a[4]*t1+a[5];
      if(f1>2000.) f1=2000.;
      if(version==2 || version == 4){
         f2 = a1*t2*t2+b1*t2+c1;
         i2 = 1;
         if(f2<0) i2= -1;
         f2 = log(fabs(f2)) + a[4]*t2+a[5];
         if(f2 > 2000.) f2=2000.;
         if(version == 2)return i1*exp(f1)-i2*exp(f2);
         return i2*exp(f2)-i1*exp(f1);
      } 
      if(version==1) return i1*exp(f1);
      return -i1*exp(f1);
   }
   a1 = (((a[1]/3)*(t2-t1)+a[2]/2)*(t2-1)+a[3])*(t2-t1)*exp(a[5]);
   if(version==4)return a1;
   return -a1;
}

/******************************************************************************/
/* This function computes a similar integral, but with a higher order leading
polinomial                                                                    */

double expin2(version,t1,t2,aa,b1,b0)

/* input: aa,b1,b0,t1,t2,version: see figure above
   local: u6,u5,u4,u3,u2,u1,u0: as a4,a3,a2,a1,a0, but for the primitive.            
          f1,f2 half-products                                    */
int version;
double t1,t2,aa[],b1,b0;
/* this version does not contain much information - the best way to figure out
   what is happening is to compute the integrals above, and then check below  */

{
   double u6,u5,u4,u3,u2,u1,u0,f1,f2;
   int i1,i2;
/* if b1=0 and version is 1 or 3 well...........                              */
   if(b1!=0 || version ==1 || version == 3){
      u6 = aa[6]/b1;
      u5 = (aa[5]-6*u6)/b1;
      u4 = (aa[4]-5*u5)/b1;
      u3 = (aa[3]-4*u4)/b1;
      u2 = (aa[2]-3*u3)/b1;
      u1 = (aa[1]-2*u2)/b1;
      u0 = (aa[0]-u1)/b1;
      f1 = (((((u6*t1+u5)*t1+u4)*t1+u3)*t1+u2)*t1+u1)*t1+u0;
      i1 = 1;
      if(f1<0) i1 = -1;
      f1 = log(fabs(f1)) + b1*t1+b0;
      if(f1>2000.) f1=2000.;
      if(version==2 || version == 4){
         f2 = (((((u6*t1+u5)*t1+u4)*t2+u3)*t2+u2)*t2+u1)*t2+u0;
         i2 = 1;
         if(f2<0) i2= -1;
         f2 = log(fabs(f2)) + b1*t2+b0;
         if(f2 > 2000.) f2=2000.;
         if(version == 2)return i1*exp(f1)-i2*exp(f2);
         return i2*exp(f2)-i1*exp(f1);
      } 
      if(version==1) return i1*exp(f1);
      return -i1*exp(f1);
   }
   u6 = (((aa[6]/7*(t2-t1)+aa[5]/6)*(t2-t1)+aa[4]/5)*(t2-t1)+aa[3]/4)*(t2-t1);
   u6 = (((u6 + aa[2]/3)*(t2-t1)+aa[1]/2)*(t2-t1)+aa[0])*(t2-t1)*exp(b0);
   if(version==4)return u6;
   return -u6;
}
/******************************************************************************/
/* this function return a value from a logspline density                      */

double dens3(x)
double x;
/* point of interest                                                          */

{
      double dens33();
      return exp(dens33(x));
}

/******************************************************************************/
/* this function return the log of a value from a logspline density           */
double dens33(x)
double x;
/* point of interest                                                          */

{
   int j,k;
   double f;

/* circle through the knots                                                   */
      for(j=0; (j<nknots && x > knots[j]); j++); 
      f = -log(ctheta);

/* find in between which knots the point is, and compute the spline           */
      for(k=j-3; k<j+1; k++){
         if(k>=0 && k<=nknots){
            f = f + theta[k] * (coef[k][0][j] + x * (coef[k][1][j] 
                  + x * (coef[k][2][j] + x * coef[k][3][j])));
         }
      }
      return f;
}
/******************************************************************************/
/* Numerical integration using gaussian quadrature.
   See Abromowitz and Stegun.

   k1 and k2: lower and upper integration bounds
   fun      : function to be integrated          */


double numint(k1,k2,fun,accuracy)

/* Intgerals using Gauss-Legendre quadrature with 12 points    
   y1,y2,... - abisces
   w1,w2,... - weight
   accuracy  - accuracy
   r1 and r2 - from (k1,k2) to (-1,1)         */

double k1,k2,(*fun)();
int accuracy;
{
   double r1,r2,w[33],y[33];
   int i;
   r1 = ((k2 - k1) / 2);
   r2 = ((k2 + k1) / 2);

   if(accuracy == 0 || accuracy == 1){
      w[1] = 0.467913934572691 * r1; y[1] = 0.238619186083197 * r1;
      w[2] = 0.360761573048139 * r1; y[2] = 0.661209386466265 * r1;
      w[3] = 0.171324429379170 * r1; y[3] = 0.932469514203152 * r1;
      r1 = 0.;
      for(i=1;i<4;i++)r1 = r1 + w[i]*(fun(r2-y[i])+fun(r2+y[i]));
      return r1;
   }

   w[ 1]=  0.00178328072169643 * r1; y[1 ]=  0.99930504173577217 * r1;
   w[ 2]=  0.00414703326056247 * r1; y[2 ]=  0.99634011677195533 * r1;
   w[ 3]=  0.00650445796897836 * r1; y[3 ]=  0.99101337147674429 * r1;
   w[ 4]=  0.00884675982636395 * r1; y[4 ]=  0.98333625388462598 * r1;
   w[ 5]=  0.01116813946013113 * r1; y[5 ]=  0.97332682778991098 * r1;
   w[ 6]=  0.01346304789671864 * r1; y[6 ]=  0.96100879965205377 * r1;
   w[ 7]=  0.01572603047602472 * r1; y[7 ]=  0.94641137485840277 * r1;
   w[ 8]=  0.01795171577569734 * r1; y[8 ]=  0.92956917213193957 * r1;
   w[ 9]=  0.02013482315353021 * r1; y[9 ]=  0.91052213707850282 * r1;
   w[10]=  0.02227017380838325 * r1; y[10]=  0.88931544599511414 * r1;
   w[11]=  0.02435270256871087 * r1; y[11]=  0.86599939815409277 * r1;
   w[12]=  0.02637746971505466 * r1; y[12]=  0.84062929625258032 * r1;
   w[13]=  0.02833967261425948 * r1; y[13]=  0.81326531512279754 * r1;
   w[14]=  0.03023465707240248 * r1; y[14]=  0.78397235894334139 * r1;
   w[15]=  0.03205792835485155 * r1; y[15]=  0.75281990726053194 * r1;
   w[16]=  0.03380516183714161 * r1; y[16]=  0.71988185017161088 * r1;
   w[17]=  0.03547221325688239 * r1; y[17]=  0.68523631305423327 * r1;
   w[18]=  0.03705512854024005 * r1; y[18]=  0.64896547125465731 * r1;
   w[19]=  0.03855015317861563 * r1; y[19]=  0.61115535517239328 * r1;
   w[20]=  0.03995374113272034 * r1; y[20]=  0.57189564620263400 * r1;
   w[21]=  0.04126256324262353 * r1; y[21]=  0.53127946401989457 * r1;
   w[22]=  0.04247351512365359 * r1; y[22]=  0.48940314570705296 * r1;
   w[23]=  0.04358372452932345 * r1; y[23]=  0.44636601725346409 * r1;
   w[24]=  0.04459055816375657 * r1; y[24]=  0.40227015796399163 * r1;
   w[25]=  0.04549162792741814 * r1; y[25]=  0.35722015833766813 * r1;
   w[26]=  0.04628479658131442 * r1; y[26]=  0.31132287199021097 * r1;
   w[27]=  0.04696818281621002 * r1; y[27]=  0.26468716220876742 * r1;
   w[28]=  0.04754016571483031 * r1; y[28]=  0.21742364374000708 * r1;
   w[29]=  0.04799938859645831 * r1; y[29]=  0.16964442042399283 * r1;
   w[30]=  0.04834476223480295 * r1; y[30]=  0.12146281929612056 * r1;
   w[31]=  0.04857546744150343 * r1; y[31]=  0.07299312178779904 * r1;
   w[32]=  0.04869095700913972 * r1; y[32]=  0.02435029266342443 * r1;
   r1 = 0.;
   for(i=1;i<33;i++)r1 = r1 + w[i]*(fun(r2-y[i])+fun(r2+y[i]));
   return r1;
}
/***************************************/
double numints(vv,k1,k2,fun,accuracy,ip)

/* Intgerals using Gauss-Legendre quadrature with 12 points    
   y1,y2,... - abisces
   w1,w2,... - weight
   accuracy  - accuracy
   r1 and r2 - from (k1,k2) to (-1,1)         */

double k1,k2,(*fun)(),vv[];
int accuracy,ip;
{
   double y[33],w[33],r1,r2;
   int i1;
   r1 = ((k2 - k1) / 2);
   r2 = ((k2 + k1) / 2);

   if(accuracy == 0 || accuracy == 1){
      w[1 ]= 0.467913934572691 * r1; y[1 ]= 0.238619186083197 * r1;
      w[2 ]= 0.360761573048139 * r1; y[2 ]= 0.661209386466265 * r1;
      w[3 ]= 0.171324429379170 * r1; y[3 ]= 0.932469514203152 * r1;
      for(i1=0;i1<ip;i1++)vv[i1]=0.;
      for(i1=1;i1<4;i1++){ 
         fun(w[i1],r2-y[i1],vv,ip);
         fun(w[i1],r2+y[i1],vv,ip);
      }
      return 0.;
   }
   w[1 ]=  0.00178328072169643 * r1; y[1 ]=  0.99930504173577217 * r1;
   w[2 ]=  0.00414703326056247 * r1; y[2 ]=  0.99634011677195533 * r1;
   w[3 ]=  0.00650445796897836 * r1; y[3 ]=  0.99101337147674429 * r1;
   w[4 ]=  0.00884675982636395 * r1; y[4 ]=  0.98333625388462598 * r1;
   w[5 ]=  0.01116813946013113 * r1; y[5 ]=  0.97332682778991098 * r1;
   w[6 ]=  0.01346304789671864 * r1; y[6 ]=  0.96100879965205377 * r1;
   w[7 ]=  0.01572603047602472 * r1; y[7 ]=  0.94641137485840277 * r1;
   w[8 ]=  0.01795171577569734 * r1; y[8 ]=  0.92956917213193957 * r1;
   w[9 ]=  0.02013482315353021 * r1; y[9 ]=  0.91052213707850282 * r1;
   w[10]=  0.02227017380838325 * r1; y[10]=  0.88931544599511414 * r1;
   w[11]=  0.02435270256871087 * r1; y[11]=  0.86599939815409277 * r1;
   w[12]=  0.02637746971505466 * r1; y[12]=  0.84062929625258032 * r1;
   w[13]=  0.02833967261425948 * r1; y[13]=  0.81326531512279754 * r1;
   w[14]=  0.03023465707240248 * r1; y[14]=  0.78397235894334139 * r1;
   w[15]=  0.03205792835485155 * r1; y[15]=  0.75281990726053194 * r1;
   w[16]=  0.03380516183714161 * r1; y[16]=  0.71988185017161088 * r1;
   w[17]=  0.03547221325688239 * r1; y[17]=  0.68523631305423327 * r1;
   w[18]=  0.03705512854024005 * r1; y[18]=  0.64896547125465731 * r1;
   w[19]=  0.03855015317861563 * r1; y[19]=  0.61115535517239328 * r1;
   w[20]=  0.03995374113272034 * r1; y[20]=  0.57189564620263400 * r1;
   w[21]=  0.04126256324262353 * r1; y[21]=  0.53127946401989457 * r1;
   w[22]=  0.04247351512365359 * r1; y[22]=  0.48940314570705296 * r1;
   w[23]=  0.04358372452932345 * r1; y[23]=  0.44636601725346409 * r1;
   w[24]=  0.04459055816375657 * r1; y[24]=  0.40227015796399163 * r1;
   w[25]=  0.04549162792741814 * r1; y[25]=  0.35722015833766813 * r1;
   w[26]=  0.04628479658131442 * r1; y[26]=  0.31132287199021097 * r1;
   w[27]=  0.04696818281621002 * r1; y[27]=  0.26468716220876742 * r1;
   w[28]=  0.04754016571483031 * r1; y[28]=  0.21742364374000708 * r1;
   w[29]=  0.04799938859645831 * r1; y[29]=  0.16964442042399283 * r1;
   w[30]=  0.04834476223480295 * r1; y[30]=  0.12146281929612056 * r1;
   w[31]=  0.04857546744150343 * r1; y[31]=  0.07299312178779904 * r1;
   w[32]=  0.04869095700913972 * r1; y[32]=  0.02435029266342443 * r1;
   for(i1=0;i1<ip;i1++)vv[i1]=0.;
   for(i1=1;i1<33;i1++){
      fun(w[i1],r2-y[i1],vv,ip);
      fun(w[i1],r2+y[i1],vv,ip);
   }
   return 0.;
}
/******************************************************************************/

/* These are 5 functions that are used in integration routines.     

   dfunpar[] - parameters used by the integrayion routines.                   */

double fun2(x)
/* This function is:
                            2      3
               p1 + x p2 + x p3 + x p4
              e
                                                                              */
double x;
{
   return exp(dfunpar[1] + x * (dfunpar[2] + x * (dfunpar[3] + x *dfunpar[4])));
}

/******************************************************************************/
double fun48(w,x,vv,ip)
/* This function is:
           p0
          x  f(x)
                 +
                      where f(x) is a density provided by dens3               */
double x,w,vv[];
int ip;
{
   double dens3(),vx;
   int i1;
   vx=w*dens3(x);
   vv[0]=vv[0]+vx;
   for(i1=1;i1<ip;i1++){
      vx=vx*x;
      vv[i1]=vv[i1]+vx;
   }
   return 0.;
}
/******************************************************************************/
