/*******************************************************************************
* Copyright (c) 1993, 1995 by Charles Kooperberg.                              *
* All rights reserved. This function is part of the Logspline Density          *
* Estimation Program. It was written by Charles Kooperberg at the University   *
* of Washington and the University of California at Berkeley between 1988 and  *
* 1991. It is described in the paper: Kooperberg, Charles and Stone, Charles J.*
* `Logspline density estimation for censored data', Journal of Computational   *
* and Graphical Statistics, 1 (1992) 301-328.                                  *
* You are free to use this program, for non-commercial purposes only,          *
* under two conditions:                                                        *
* (1) This note is not to be removed                                           *
* (2) Publications using logspline computations should reference the           *
*  publication  mentioned above.                                               *
* For questions, please email clk@stat.washington.edu                          *
*                       Charles Kooperberg, November 30, 1995                  *
*******************************************************************************/
/* this file contains a few miscelaneous routines

   erroradjust - computes the error criterion
   likeli      - computes the loglikelihood
   linsearch   - computes the winning stepsize for steepest decent
   onesearch   - computes ctheta and the loglikelihood for one theta+shift    */

/******************************************************************************/
#include "mhead.h"
#include <math.h>

/* this function computes the stopcriterion (error) and adjust the stepsize 
   (shift) if this is too large.                                              */

double erroradjust(shift)
double shift[];

{
   double r1,r2;
   int i,numbertester();
/* all utility numbers */

/* the error is the sum of (shift/theta)^2, except where theta is very small  */
   r1 = 0.;
   for(i=0; i<nknots-1; i++){
      r2 = theta[i] * theta[i];
      if(r2 < 0.0000001) r2 = 0.0000001;
      r1 = r1 + shift[i] * shift[i] / r2;
   }
   if(numbertester(r1)!=1)r2 = sqrt(r1);

/* if the error is infinity we compute an alternate error criterion to prevent
   trouble                                                                    */
   else{
      r1 = 0.;
      for(i=0; i<nknots-1; i++){
         r2 = fabs(theta[i]);
         if(r2<0.00001) r2 = 0.00001;
         r1 = r1 + fabs(shift[i]) / r2; 
      }
      r2=r1;
   }

/* If the error would be too big, we only take a smaller step: this
   is to prevent overshoot.                                                   */
   for(i=0;i<nknots-1 && r1 > 1000.;i++) shift[i]=shift[i] * 3. / r2;

   return r1;
}
double error2(shift,rvr)
double shift[],rvr[];
{
   int i,j,numbertester();
   double r=0.;
   for(i=0;i<nknots-1;i++)r=r+shift[i]*rvr[i]/2.;
   j=-1;
   if(r>=0.0)j++;
   if(numbertester(r)==1)j=j-2;
   if(j<0)r=1000;
   return r;
}
/******************************************************************************/

/* this routine computes the loglikelihood                                    */

double likeli(candidate,nsample,sample,bound,accuracy)
double candidate[],sample[],bound[];
int nsample[],accuracy;
{
   double r0,r1,likl,dens3(),r3[NC+1],aa[6],numint(),expin(),dens33(),bb[6];
   int i1,i2,i3,i4,iv,iw;
   r0=exp((double)(-740));

/* the function uses numint and dens33 - which always use theta, but the
   routine doesn't always compute the likelihood in theta, sometimes at
   another place - so we have to swap theta and candidate                     */
   for(i1=0;i1<(nknots-1);i1++){
      r1=theta[i1];
      theta[i1]=candidate[i1];
      candidate[i1]=r1;
   }

/* the stuff for the exact data is easy                                       */
   likl=0.;
   for(i1=0;i1<nsample[1];i1++)
         likl=likl+dens33(sample[i1]);

/* compute the integral of the density between each two knots in r3           */
   if(bound[3]>0.5)iv=4;
   else iv=3;
   if(bound[1]>0.5)iw=2;
   else iw=1;
   aa[1]=0.;
   aa[2]=0.;
   aa[3]=1.;
   aa[4]=theta[nknots-2]*coef[nknots-2][1][nknots];
   aa[5]=theta[nknots-2]*coef[nknots-2][0][nknots]
     +theta[nknots-3]*coef[nknots-3][0][nknots]-log(ctheta);
   if(nknots<4)
   aa[5]=theta[nknots-2]*coef[nknots-2][0][nknots]-log(ctheta);
   r3[nknots]=expin(iv,knots[nknots-1],bound[4],aa);
   bb[1]=0.;
   bb[2]=0.;
   bb[3]=1.;
   bb[4]=theta[0] * coef[0][1][0];
   bb[5]=theta[0] * coef[0][0][0]-log(ctheta);
   r3[0]=expin(iw,knots[0],bound[2],bb);
   for(i1=1;i1<nknots;i1++)
      r3[i1]=numint(knots[i1-1],knots[i1],dens3,accuracy);

/* depending on the accuracy we either deal with the censored data approximate
   or exact                                                                   */
   if(accuracy!=0){
/* exact first the rounded data                                               */
      for(i1=0;i1<nsample[2];i1++){
         i2=i1*2+nsample[1];
/* between which knots are the two endpoints                                  */
         for(i3=0;knots[i3]<sample[i2] && i3 < nknots;i3++);
         for(i4=0;knots[i4]<sample[i2+1] && i4 < nknots;i4++);
/* is it the same interval                                                    */
         if(i3==i4)
            likl=likl+log(numint(sample[i2],sample[i2+1],dens3,0));
/* or a different interval                                                    */
         else{
            r1=numint(sample[i2],knots[i3],dens3,0);
            r1=r1+numint(knots[i4-1],sample[i2+1],dens3,0);
            for(i3=i3+1;i3<i4;i3++)r1=r1+r3[i3];
            likl=likl+log(r1);
         }
      }
/* the right censored data                                                    */
      for(i1=0;i1<nsample[3];i1++){
         i2=i1+nsample[1]+2*nsample[2];
/* in which interval is the datapoint                                         */
         for(i3=0;knots[i3]<sample[i2] && i3 < nknots;i3++);
/* somewhere in the middle                                                    */
         if(i3<nknots){
            r1=numint(sample[i2],knots[i3],dens3,0);
            for(i3=i3+1;i3<=nknots;i3++) r1=r1+r3[i3];
/* if r1<=r0 the likl is effectively -infinity                                */
            if(r1>r0) likl=likl+log(r1);
            else likl=likl-1000;
         }
/* in the tail                                                                */
         else likl=likl+log(expin(iv,sample[i2],bound[4],aa));
      }
/* the left censored data                                                    */
      for(i1=0;i1<nsample[4];i1++){
         i2=i1+nsample[1]+2*nsample[2]+nsample[3];
/* in which interval is the datapoint                                         */
         for(i3=0;knots[i3]<sample[i2] && i3 < nknots;i3++);
/* somewhere in the middle                                                    */
         if(i3>0){
            r1=numint(knots[i3-1],sample[i2],dens3,0);
            for(i4=0;i4<i3;i4++) r1=r1+r3[i4];
/* if r1<=r0 the likl is effectively -infinity                                */
            if(r1>r0) likl=likl+log(r1);
            else likl=likl-1000;
         }
/* in the tail                                                                */
         else likl=likl+log(expin(iw,sample[i2],bound[2],bb));
      }
   }
   else{
/* approximate, first right censored - essentially as above   
   there are ng3 points at xg                                                 */
      if(nsample[3]>0){
         i1=0;
         for(i2=0;i2<NC;i2++){
            if(ng3[i2]>0){
               i1=i1+ng3[i2];
               for(i3=0;knots[i3]<xg[i2] && i3<nknots;i3++);
               if(i3<nknots){
                  r1=numint(xg[i2],knots[i3],dens3,0);
                  for(i3=i3+1;i3<=nknots;i3++) r1=r1+r3[i3];
                  if(r1>r0) likl=likl+ng3[i2]*log(r1);
                  else likl=likl-1000.*ng3[i2];
               }
               else likl=likl+ng3[i2]*log(expin(iv,xg[i2],bound[4],aa));
               if(i1==nsample[3])i2=NC+3;
            }   
         }   
      }   
/* approximate, now left censored - essentially as above   
   there are ng4 points at xg                                                 */
      if(nsample[4]>0){
         i1=0;
         for(i2=0;i2<NC;i2++){
            if(ng4[i2]>0){
               i1=i1+ng4[i2];
               for(i3=0;knots[i3]<xg[i2] && i3<nknots;i3++);
               if(i3>0){
                  r1=numint(knots[i3-1],xg[i2],dens3,0);
                  for(i4=0;i4<i3;i4++) r1=r1+r3[i4];
                  if(r1>r0) likl=likl+ng4[i2]*log(r1);
                  else likl=likl-1000.*ng4[i2];
               }
               else likl=likl+ng4[i2]*log(expin(iw,xg[i2],bound[2],bb));
               if(i1==nsample[4])i2=NC+3;
            }   
         }   
      }   
/* the interval censored stuff is easy                                        */
      if(nsample[2]>0){
         for(i1=0;i1<nsample[2];i1++){
            r1=(sample[nsample[1]+i1*2]+sample[nsample[1]+i1*2+1])/2.;
            likl=likl+dens33(r1);
         }
      }
   }
/* swap theta and candidate back                                              */
   for(i1=0;i1<(nknots-1);i1++){
      r1=theta[i1];
      theta[i1]=candidate[i1];
      candidate[i1]=r1;
   }
   return likl/(double)nsample[0];
}
/******************************************************************************/

/* this routine does a steepest decent search in the direction shift          */

double linsearch(shift,oldll,bound,nsample,sample,accuracy)
double shift[],oldll,bound[],sample[];
int nsample[],accuracy;
{
   double rt,rl=1000.,rr=1000.,maxll,maxrt,ll,onesearch();
   int i,err;
   if(bound[1]<0.5 && shift[0]*theta[0]<0)rl=-theta[0]/shift[0];
   if(bound[3]<0.5 && shift[nknots-2]*theta[nknots-2]<0)
                           rr=-theta[nknots-2]/shift[nknots-2];
/* rl and rr are the maximum stepsize, given the sign of theta and integration
   boundaries                                                                 */
   if(rl<rr)rr=rl;
   maxll=oldll;
   maxrt=-100;
/* rt is the first stepsize guess, on a 2log basis                            */
   if(rr>0.5)rt=-2.;
   else rt=(double)floor(log(rr)/log((double)2))-2.;
/* onesearch computes the loglikelihood for one rt (stepsize)                 */
   ll=onesearch(rt,shift,accuracy,bound,&err,nsample,sample);
/* if for the basis stepsize the likelihood goes up we try more               */
   if(err==0 && ll>maxll){
      do{
         maxrt=rt;
         maxll=ll;
         rt=rt+2.;
         ll=onesearch(rt,shift,accuracy,bound,&err,nsample,sample);
      }while(rt<8.5 && ll>maxll && err==0);
   }
/* if for the basis stepsize the likelihood goes down we try less             */
   else {
      do{
         rt=rt-2.;
         ll=onesearch(rt,shift,accuracy,bound,&err,nsample,sample);
      }while(rt>-14.1 && (ll < maxll || err==1));
   }
/* write down the winning combination                                         */
   if(err==0 && ll>maxll){
      maxrt=rt;
      maxll=ll;
   }
   if(maxrt > -50.){
      rt = maxrt + 1;
      ll=onesearch(rt,shift,accuracy,bound,&err,nsample,sample);
      if(err==0 && ll>maxll){
         maxrt=rt;
         maxll=ll;
      }
      maxrt=pow(2.,maxrt);
      for(i=0;i<nknots-1;i++)theta[i]=theta[i]+maxrt*shift[i];
   }
   return maxll;
}
   
/******************************************************************************/

/* this routine computes the loglikelihood in theta+2^rt shift                */

double onesearch(rt,shift,accuracy,bound,err,nsample,sample)
double shift[],rt,bound[],sample[];
int accuracy,*err,nsample[];
{
   int i,j;
   double rv[NC],middle(),likeli(),ll,dm1[NC][NC],dm2[NC],dm3[NC][2];
   *err=0;
   rt=pow(2.,rt);
   for(i=0;i<nknots-1;i++)rv[i]=theta[i]+rt*shift[i];
/* compute ctheta                                                             */
   ctheta=middle(dm1,dm2,dm3,bound,accuracy,nsample,sample,rv,0);
/* error conditions                                                           */
   if(ctheta<=0){
      *err=1;
      return 0.;
   }
   if(log(ctheta)>200){
      *err=1;
      return 0.;
   }
   j=0;
   if(ctheta > -5)j++; 
   if(ctheta < 5)j++; 
   if(j==0){
      *err=1;
      return 0.;
   }
/* compute likeli                                                             */
   ll=likeli(rv,nsample,sample,bound,accuracy);
   return ll;
}
int numbertester(aa)
double aa;
/* if aa = -Inf: 0
      aa = +Inf: 1
      aa =  NaN: 2
      otherwise: 3 */
{
   int i1=0,i2=0,i3=0,i4=0;
   if(aa< 2.){
      i1=1;
   }
   if(aa> 0.){
      i2=1;
   }
   if(aa< pow(10.,200.)){
      i3=1;
   }
   if(aa> -pow(10.,200.)){
      i4=1;
   }
   if(i1+i2+i3+i4>=3){
      return 3;
   }
   if(i2==1 && i4==1){
      return 1;
   }
   if(i1==1 && i3==1){
      return 0;
   }
   return 2;
}
